
/*
 *
 *    Copyright  2008 Sun Microsystems, Inc. All rights reserved
 *
 *    Use is subject to license terms.
 *
 */

import javax.realtime.Clock;
import javax.realtime.AbsoluteTime;
import javax.realtime.RelativeTime;
import javax.realtime.SchedulingParameters;
import javax.realtime.PriorityParameters;
import javax.realtime.ReleaseParameters;
import javax.realtime.PeriodicParameters;
import javax.realtime.PriorityScheduler;
import javax.realtime.RealtimeThread;
import javax.realtime.Scheduler;

public class GCDeterministic implements Runnable {
    
static final int NB_CHECK_OUTER_ITERATIONS = 100;



    static {
        System.out.println("\n--------------- Running GCDeterministic Program ---------------");
        System.out.println("\nNote:\n    All time measurements are expressed in microsecond units.");

        // Forces a load and initialization of Fibonacci classes
        Fibonacci1 f1 = new Fibonacci1(10);
        Fibonacci2 f2 = new Fibonacci2(10);
        f1.computeFib();
        f2.computeFib();
        
        // Forces a load and initialization of GarbageProducer classes
        GarbageProducer1 gp1 = new GarbageProducer1(10);
        GarbageProducer2 gp2 = new GarbageProducer2(10);
        gp1.produceGarbage(10);
        gp2.produceGarbage(10);

        // Forces a load and initialization of plugged in code
        // Uncomment the next 3 lines, and replace Fibonacci1 by the name
        // of your own StressingObject class
        // to preload and preinitialize your own stressing code.

        // GCDeterministic tmpGCDeterministic = new GCDeterministic(1);
        // tmpGCDeterministic.setVerbosity(false);
        // tmpGCDeterministic.stress("Fibonacci1", 10, 10);

        System.out.println("Static initialization finished:");
        System.out.println("  ==> Stressing Fibonacci and GarbageProducer classes initialized and loaded");
    }
    
    // stressing iteration data ( captured as program parameters)
    int nb_outer_iterations = 0;
    int nb_inner_iterations = 0;
    int nb_stress_class_iterations = 0;
    int pause_time = 0;

    // stressing Allocation data ( captured as program parameters )
    int allocation_array_size = 0;
    boolean produce_garbage = false;

    // verbosity flag (captured as program parameter)
    boolean verbosity = false;
    
    // stress java.lang.Thread priority
    int thread_priority;

    // timestamps array for bench thread
    long[] wallclock_times_begin_loops = null; 
    long[] wallclock_times_end_loops = null; 

    // array of number of missed deadline fro a given loop
    int[] missed_deadlines = null;

    // measurements instance
    Measurements measurements = null;
   
    // delegates result printing to measurement
    private void outputResults() {
        measurements.computeAndPrintStatResults(wallclock_times_begin_loops, wallclock_times_end_loops, missed_deadlines);
    }

    /** Instantiate a StressingObject from stressingObjectClassName and
        and stresses nb_inner_iterations times with this stressingObject
        which is supposed to iterate nb_stress_class_iterations times by itself */
    void stress(String stressingObjectClassName, int nb_inner_iterations, int nb_stress_class_iterations) {
        StressingObject so = null;
        
        try {
            so = (StressingObject) (Class.forName(stressingObjectClassName).newInstance());
            so.setStressClassIterations(nb_stress_class_iterations);
            for (int i = 1; i <= nb_inner_iterations; i++) {
                so.stress();
            }
        }
        catch(Exception e) {System.out.println(e);}
    }

    // Fibonacci computation stress method
    // computes 2 fibonacci sequences "nb_inner_iterations" times
    //  -  nb_stress_class_iterations : quantity of fib numbers to compute
    void computeFibs(int nb_inner_iterations, int nb_stress_class_iterations) {
        Fibonacci1 f1 = new Fibonacci1(nb_stress_class_iterations);
        Fibonacci2 f2 = new Fibonacci2(nb_stress_class_iterations);
        
        for (int i = 1; i <= nb_inner_iterations; i++) {
            f1.computeFib();
            f2.computeFib();
        }
    }

    // Memory allocation stress method
    // 2 garbage production sequences "nb_inner_iterations" times
    //  - nb_stress_class_iterations : number of int arrays to allocate
    //  - allocation_array_size: size of the arrays to allocate
    void produceGarbage(int nb_inner_iterations, int nb_stress_class_iterations, int allocation_array_size) {

        GarbageProducer1 gp1 = new GarbageProducer1(allocation_array_size);
        GarbageProducer2 gp2 = new GarbageProducer2(allocation_array_size);
        
        for (int i = 1; i <= nb_inner_iterations; i++) {
            gp1.produceGarbage(nb_stress_class_iterations);
            gp2.produceGarbage(nb_stress_class_iterations);
        }
    }

    /** Creates a new instance of Deterministic */
    GCDeterministic(int nb_loops) {
        wallclock_times_begin_loops = new long[nb_loops];
        wallclock_times_end_loops = new long[nb_loops];
        missed_deadlines = new int[nb_loops];
        
        measurements = new Measurements(nb_loops);
    }
    

    /** Changes verbosity level */
    private void setVerbosity(boolean verbose_flag) {
        verbosity = verbose_flag;
        measurements.setVerbosity(verbosity);
    }

    // For additional time consumption from a time-sharing thread
    public void run() {
        try {
            for (int i = 0; i < this.nb_outer_iterations; i++) {
                // fibonacci calculation
                computeFibs(this.nb_inner_iterations, this.nb_stress_class_iterations);

                // Stressing with your own code.
                // Replace Fibonacci1 by your StressingObject class name
                // and uncomment the following line to stress with your own code.
                // stress("Fibonacci1", nb_inner_iterations, nb_stress_class_iterations);
                
                // parameterized pause time
                if (this.pause_time != 0) Thread.currentThread().sleep((long) this.pause_time);

                // parameterized garbage memory production
                if (this.produce_garbage) produceGarbage(this.nb_inner_iterations, this.nb_stress_class_iterations, this.allocation_array_size);
            }
        }
        catch (InterruptedException ie) {
            System.out.println("java.lang stress thread interrupted");
        }
    }    

    // Dedicated inner class to evaluate to execution time needed for each loop
    class CheckMeanCalculationForGarbageAndFibonacciLoops extends RealtimeThread {
        long meantime_to_compute = 0;
        int nb_inner_iterations =0;
        int nb_stress_class_iterations = 0;
        int check_allocation_array_size = 0;
        boolean garbage_producer = false;

        CheckMeanCalculationForGarbageAndFibonacciLoops(
                               int inner_iterations,
                               int stress_class_iterations,
                               int alloc_array_size,
                               SchedulingParameters schedulingParams) {
            super(schedulingParams);
            nb_inner_iterations = inner_iterations;
            nb_stress_class_iterations = stress_class_iterations;
            check_allocation_array_size = alloc_array_size;
            if (check_allocation_array_size != 0) garbage_producer = true;
        }

        // Returns the mean time needed to perform the computation given the input parameters
        // This must be used after the thread has been run once
        public RelativeTime getComputingMeanTime() {
            return new RelativeTime(meantime_to_compute / 1000000L,
                                    (int) (meantime_to_compute % 1000000L));
        }

        public void run() {
                
            final Clock rtClock = Clock.getRealtimeClock();
            int no_loop = 0;

            // the AbsoluteTime from which to measure
            AbsoluteTime rtWallClockTimeBefore = rtClock.getTime();
            AbsoluteTime rtWallClockTimeAfter = rtClock.getTime();


              // recording time before computation and allocation 
              rtClock.getTime(rtWallClockTimeBefore);

              long wallclock_times_begin_loops =
                     (rtWallClockTimeBefore.getMilliseconds() * 1000000L) +
                      rtWallClockTimeBefore.getNanoseconds();

              while (no_loop < NB_CHECK_OUTER_ITERATIONS) {
                // Fibonacci computation sample
                GCDeterministic.this.computeFibs(nb_inner_iterations, nb_stress_class_iterations);

                if (garbage_producer) {
                    GCDeterministic.this.produceGarbage(nb_inner_iterations, nb_stress_class_iterations, check_allocation_array_size);
                }

                no_loop++;
              }

              // recording time after computation and allocation 
              rtClock.getTime(rtWallClockTimeAfter);
              long wallclock_times_end_loops =
                      (rtWallClockTimeAfter.getMilliseconds() * 1000000L) +
                       rtWallClockTimeAfter.getNanoseconds();

              meantime_to_compute = 
                (wallclock_times_end_loops - wallclock_times_begin_loops) / NB_CHECK_OUTER_ITERATIONS;

        }
    }


    class RealTimeGarbageLoops extends RealtimeThread {

        boolean records_time_measures = false;
        int nb_outer_iterations = 0;
        int nb_inner_iterations = 0;
        int nb_stress_class_iterations = 0;
        int allocation_array_size = 0;
        boolean garbage_producer = false;
        int pause_time = 0;

        RealTimeGarbageLoops(boolean time_recording_state,
                             int outer_iterations,
                             int inner_iterations,
                             int stress_class_iterations,
                             int alloc_array_size,
                             int time_to_pause,
                             SchedulingParameters schedulingParams,
                             ReleaseParameters releaseParams) {
            super(schedulingParams,releaseParams);
            records_time_measures = time_recording_state;
            nb_outer_iterations = outer_iterations;
            nb_inner_iterations = inner_iterations;
            nb_stress_class_iterations = stress_class_iterations;
            allocation_array_size = alloc_array_size;
            if (allocation_array_size != 0) garbage_producer = true;
            pause_time = time_to_pause;
            
        }


        public void run() {
                
            
            final Clock rtClock = Clock.getRealtimeClock();
            final RelativeTime time_to_sleep = new RelativeTime(pause_time, 0);


            // the AbsoluteTime from which to measure
            AbsoluteTime rtWallClockTimeBefore = rtClock.getTime();
            AbsoluteTime rtWallClockTimeAfter = rtClock.getTime();

            int no_loop = 0;
            try {

                while (no_loop < nb_outer_iterations) {

                    if (records_time_measures) {
                        // bench thread -> recording time before computation and allocation 
                        rtClock.getTime(rtWallClockTimeBefore);
                        wallclock_times_begin_loops[no_loop] =
                            (rtWallClockTimeBefore.getMilliseconds() * 1000000L) +
                            rtWallClockTimeBefore.getNanoseconds();
                    }

                    // fibonacci computation sample
                    GCDeterministic.this.computeFibs(nb_inner_iterations, nb_stress_class_iterations);

                    // stressing with your own code.
                    // Replace Fibonacci1 by your StressingObject class name
                    // and uncomment the following line to stress with your own code.
                    // GCDeterministic.this.stress("Fibonacci1", nb_inner_iterations, nb_stress_class_iterations);
                
                    // allocation  sample
                    if (garbage_producer) {
                        GCDeterministic.this.produceGarbage(nb_inner_iterations, nb_stress_class_iterations, allocation_array_size);
                    }

                    if (records_time_measures) {
                        // this should be bench thread -> recording time after computation and allocation 
                        rtClock.getTime(rtWallClockTimeAfter);
                        wallclock_times_end_loops[no_loop] =
                            (rtWallClockTimeAfter.getMilliseconds() * 1000000L) +
                            rtWallClockTimeAfter.getNanoseconds();
                                                
                        while(RealtimeThread.currentRealtimeThread().waitForNextPeriod() == false ) {
                            missed_deadlines[no_loop]++;
                        }
                    }
                    else {
                        // stress thread -> parameterized pause if asked for
                        if (pause_time != 0) sleep(time_to_sleep);
                    }

                    no_loop++;
                }
            }
            catch (InterruptedException ie) {
                System.out.println("javax.RealtimeThread interrupted");
            }
        }

    }




    /**
     * @param args the command line arguments
     */
    public static void main(String[] args) {
        
        // local thread variables
        Thread jlt_stress_thread = null;
        RealTimeGarbageLoops rt_bench_thread = null;
        RealTimeGarbageLoops rt_stress_thread = null;

        int jlt_stress_thread_priority = 0;
        int rt_bench_thread_priority = 0;
        int rt_stress_thread_priority = 0;

        // local run variables
        boolean jlt_stress_thread_present = false;
        boolean jrt_stress_thread_present = false;

        // extract program command-line parameters 
           // The javax.RealtimeThread bench thread parameters (all mandatory - no value check)
           int bench_nb_outer_iterations = new Integer(args[0]).intValue();
           int bench_nb_inner_iterations = new Integer(args[1]).intValue();
           int bench_nb_stress_class_iterations = new Integer(args[2]).intValue();
           int bench_allocation_array_size  = new Integer(args[3]).intValue();
           int bench_pause_time  = new Integer(args[4]).intValue();
           String bench_priority_level = args[5];

           // The javax.RealtimeThread stress thread parameters (all mandatory - no value check)
           int jrt_stress_nb_outer_iterations = new Integer(args[6]).intValue();
           int jrt_stress_nb_inner_iterations = new Integer(args[7]).intValue();
           int jrt_stress_nb_stress_class_iterations = new Integer(args[8]).intValue();
           int jrt_stress_allocation_array_size  = new Integer(args[9]).intValue();
           int jrt_stress_pause_time  = new Integer(args[10]).intValue();
           String jrt_stress_priority_level = args[11];

           // The java.lang.Thread stress thread parameters (all mandatory - no value check)
           int jlt_stress_nb_outer_iterations = new Integer(args[12]).intValue();
           int jlt_stress_nb_inner_iterations = new Integer(args[13]).intValue();
           int jlt_stress_nb_stress_class_iterations = new Integer(args[14]).intValue();
           int jlt_stress_allocation_array_size  = new Integer(args[15]).intValue();
           int jlt_stress_pause_time  = new Integer(args[16]).intValue();
           String jlt_stress_priority_level = args[17];


        // Initialization   
        PriorityScheduler ps = (PriorityScheduler)(Scheduler.getDefaultScheduler());
        final GCDeterministic gcd = new GCDeterministic(bench_nb_outer_iterations);
        

        // verbosity level (optional)
        if ((args.length >= 16) && (args[15].equals("verbose"))) gcd.setVerbosity(true);

        // check if stress thread will be present
        if ((jrt_stress_nb_outer_iterations != 0) && 
            (jrt_stress_nb_inner_iterations != 0) &&
            (jrt_stress_nb_stress_class_iterations != 0))
            jrt_stress_thread_present = true;
        else jrt_stress_thread_present = false;


        // compute average computation time for bench thread
        CheckMeanCalculationForGarbageAndFibonacciLoops computeBenchThreadTime =
            gcd.new CheckMeanCalculationForGarbageAndFibonacciLoops(
                    bench_nb_inner_iterations,
                    bench_nb_stress_class_iterations,
                    bench_allocation_array_size,
                    new PriorityParameters(ps.getMaxPriority()));

        System.out.println("\nComputing stress cost for bench thread");
        computeBenchThreadTime.start();
        try {
            computeBenchThreadTime.join(); 
        }
        catch (InterruptedException ie) {
            System.out.println("Stress cost thread for bench thread was interrupted:\n" + ie );
            System.exit(1);
        }
        RelativeTime benchThreadComputationTime = computeBenchThreadTime.getComputingMeanTime();

        // period and deadline for bench thread
        System.out.println("  Measured computation time for bench: " + benchThreadComputationTime);
        RelativeTime benchPeriod = benchThreadComputationTime.add(new RelativeTime(bench_pause_time,0));
        if (bench_pause_time == 0) benchPeriod = benchPeriod.add(new RelativeTime(0,100000));
        System.out.println("  Computed period for bench: " + benchPeriod);
        System.out.println("Stress cost time for bench computed\n");

        // Bench realtime thread instanciation
        // Iteration values must all be set to strictly positive for bench threadR
        if ((bench_nb_outer_iterations <= 0) || 
            (bench_nb_inner_iterations <= 0) ||
            (bench_nb_stress_class_iterations <= 0)) {
            System.out.println("invalid number of iterations specified for bench thread");
            System.exit(1);
        }
        else {
            if (bench_priority_level.equalsIgnoreCase("max"))
               rt_bench_thread_priority = ps.getMaxPriority();
            else if (bench_priority_level.equalsIgnoreCase("norm"))
               rt_bench_thread_priority = ps.getNormPriority();
            else rt_bench_thread_priority = ps.getMinPriority();
            
            rt_bench_thread =
                gcd.new RealTimeGarbageLoops(
                   true,
                   bench_nb_outer_iterations,
                   bench_nb_inner_iterations,
                   bench_nb_stress_class_iterations,
                   bench_allocation_array_size,        
                   bench_pause_time,
                   new PriorityParameters(rt_bench_thread_priority),
                   new PeriodicParameters(null,
                                          benchPeriod, null, null,
                                          null, null ));
        }


        // Realtime stress thread instanciation
        if (jrt_stress_thread_present) {

            // negative iteration values are ignored and changed to positive
            if ((jrt_stress_nb_outer_iterations < 0) || 
                (jrt_stress_nb_inner_iterations < 0) ||
                (jrt_stress_nb_stress_class_iterations < 0)) {
                System.out.println("negative iteration values changed to positive for javax.realtime.RealtimeThread stress thread");
                if (jrt_stress_nb_outer_iterations < 0)
                   jrt_stress_nb_outer_iterations = - jrt_stress_nb_outer_iterations;
                if (jrt_stress_nb_inner_iterations < 0)
                   jrt_stress_nb_inner_iterations = - jrt_stress_nb_inner_iterations;
                if (jrt_stress_nb_stress_class_iterations < 0)
                   jrt_stress_nb_stress_class_iterations = - jrt_stress_nb_stress_class_iterations;
            }

            // setting requested priority for stress Realtime thread
            if (jrt_stress_priority_level.equalsIgnoreCase("max"))
               rt_stress_thread_priority = ps.getMaxPriority();
            else if (jrt_stress_priority_level.equalsIgnoreCase("norm"))
               rt_stress_thread_priority = ps.getNormPriority();
            else rt_stress_thread_priority = ps.getMinPriority();

            rt_stress_thread =
                gcd.new RealTimeGarbageLoops(
                   false,
                   jrt_stress_nb_outer_iterations,
                   jrt_stress_nb_inner_iterations,
                   jrt_stress_nb_stress_class_iterations,
                   jrt_stress_allocation_array_size,        
                   jrt_stress_pause_time,
                   new PriorityParameters(rt_stress_thread_priority),
                   null);
        }
        else jrt_stress_thread_present = false;

        // java.lang.Thread stress thread instanciation
        // Any iteration value to zero means NO java.lang.Thread stress thread 
        if ((jlt_stress_nb_outer_iterations != 0) &&
            (jlt_stress_nb_inner_iterations != 0) &&
            (jlt_stress_nb_stress_class_iterations != 0)) {

            jlt_stress_thread_present = true;

            // Setting priority for stress java.lang.Thread
            if (jlt_stress_priority_level.equalsIgnoreCase("max"))
               jlt_stress_thread_priority = Thread.MAX_PRIORITY;
            else if (jlt_stress_priority_level.equalsIgnoreCase("norm"))
               jlt_stress_thread_priority = Thread.NORM_PRIORITY;
            else jlt_stress_thread_priority = Thread.MIN_PRIORITY;

            jlt_stress_thread = new Thread(gcd);
            jlt_stress_thread.setPriority(jlt_stress_thread_priority);
        }
        else {
            jlt_stress_thread_present = false;
        }

        // Iteration data ( captured as program parameters)
        gcd.nb_outer_iterations = jlt_stress_nb_outer_iterations;
        gcd.nb_inner_iterations = jlt_stress_nb_inner_iterations;
        gcd.nb_stress_class_iterations = jlt_stress_nb_stress_class_iterations;
        gcd.pause_time = jlt_stress_pause_time;
        gcd.allocation_array_size = jlt_stress_allocation_array_size;
        if (gcd.allocation_array_size != 0) gcd.produce_garbage = true;
        gcd.thread_priority = jlt_stress_thread_priority;

        // verbosity level (optional)
        if ((args.length >= 19) && (args[18].equals("verbose"))) gcd.setVerbosity(true);

        System.out.println("   Initialization and configuration of program finished ");

        // recap of options 
        System.out.println("\n============================================");
        System.out.println("   GCDeterministic program parameters");
        System.out.println("============================================");
        System.out.println("   The options used for this run are:   ");
        System.out.println("----------------------------------------");
        System.out.println("  Realtime bench thread parameters:");
        System.out.println("  =================================");
        System.out.println("   Number of outer loops: " + bench_nb_outer_iterations);
        System.out.println("   Number of inner iterations: " + bench_nb_inner_iterations);
        System.out.println("   Number of stress classes iterations: " + bench_nb_stress_class_iterations);
        System.out.println("   Size of elementary array to allocate: " + bench_allocation_array_size);
        System.out.println("   Entered pause time: " + bench_pause_time + " milliseconds");
        System.out.println("   Calculated period (deadline + pause): " + benchPeriod);
        System.out.println("   Priority level: " + bench_priority_level);
        System.out.println("   Effective priority assigned: " + rt_bench_thread_priority);
        System.out.println();

        if (jrt_stress_thread_present) {
            System.out.println("  Realtime stress thread parameters:");
            System.out.println("  =================================");
            System.out.println("   Number of outer loops: " + jrt_stress_nb_outer_iterations);
            System.out.println("   Number of inner iterations: " + jrt_stress_nb_inner_iterations);
            System.out.println("   Number of stress classes iterations: " + jrt_stress_nb_stress_class_iterations);
            System.out.println("   Size of elementary array to allocate: " + jrt_stress_allocation_array_size);
            System.out.println("   Entered pause time: " + jrt_stress_pause_time + " milliseconds");
            System.out.println("   Priority level: " + jrt_stress_priority_level);
            System.out.println("   Effective priority assigned: " + rt_stress_thread_priority);
            System.out.println();
        }

        if (jlt_stress_thread_present) {
            System.out.println("  java.lang.Thread stress thread parameters:");
            System.out.println("  ==========================================");
            System.out.println("   Number of outer loops: " + gcd.nb_outer_iterations);
            System.out.println("   Number of inner iterations: " + gcd.nb_inner_iterations);
            System.out.println("   Number of stress classes iterations: " + gcd.nb_stress_class_iterations);
            System.out.println("   Size of elementary array to allocate: " + gcd.allocation_array_size);
            System.out.println("   Inner pause time: " + gcd.pause_time);
            System.out.println("   Priority level: " + jlt_stress_priority_level);
            System.out.println("   Effective priority assigned: " + gcd.thread_priority);
            System.out.println();
        }
        
        System.out.println("  Output detail level: " + (gcd.verbosity?"detailed":"summary"));
        System.out.println("----------------------------------------\n");

        System.out.println("\n ==================== Starting " + bench_nb_outer_iterations + " iterations ===================================== \n ");


        // stabilize memory and output state before allocating objects
        System.gc();
        System.out.println("Total amount of VM memory : " + Runtime.getRuntime().totalMemory());
        System.gc();
        System.out.println("Amount of free  VM memory : " + Runtime.getRuntime().freeMemory());
        System.gc();
        System.out.println("Max amount of   VM memory : " + Runtime.getRuntime().maxMemory());
        

        // starts the java.lang.Thread stress thread
        if (jlt_stress_thread_present) {
            System.out.println("   Starting java.lang.Thread stress thread");
            jlt_stress_thread.start();
        }

        // starts the javx.RealtimeThread bench thread
        System.out.println("   Starting javax.RealtimeThread bench thread");
        rt_bench_thread.start();

        // starts the javax.RealtimeThread stress thread
        if (jrt_stress_thread_present) {
            System.out.println("   Starting javax.RealtimeThread stress thread");
            rt_stress_thread.start();
        }

        try {
            rt_bench_thread.join();
            System.out.println("   Bench javax.RealtimeThread terminated ");

            if (jrt_stress_thread_present) {
                rt_stress_thread.join();
                System.out.println("   Stress javax.RealtimeThread terminated");
            }
            
            if (jlt_stress_thread_present) {
                jlt_stress_thread.join();
                System.out.println("   Stress java.lang.Thread terminated");
            }                

            System.out.println("\n ==================== Requested " + bench_nb_outer_iterations + " iterations finished ===================================== \n");

            System.out.println("\n ==================== Starting output results ==================== \n");
            gcd.outputResults();
            System.out.println("\n  ==================== Finished results output ==================== \n ");

        
        }
        catch (InterruptedException ie) {
            System.out.println(ie);
        }
        
    }  
}

