
/*
 *
 *    Copyright  2008 Sun Microsystems, Inc. All rights reserved
 *
 *    Use is subject to license terms.
 *
 */

import javax.realtime.AbsoluteTime;


public class Measurements {

    static final int TIME_UNIT_OF_MICROSECONDS = 1000; // converting constant

    // verbosity flag :
    //  - false ==> summary of results only
    //  - true  ==> full results
    boolean verbosity = false;

    // number of loops for which to compute measures
    int nb_loops = 0;

    // Measurement and recording structures for time data
       // wallclock time 
       long[] exec_time_array;
       long[] mean_exec_time_array;
       
       // deltas to mean and ref values
       int[]  delta_to_mean_exec_time_array;
       int[]  delta_to_ref_time_array;

       // deltas to previous exec time
       int[]  delta_to_previous_array;
       int[]  mean_delta_to_previous_array;

       // accumulated minimum and maximum values
       long[] min_exec_time_array;
       long[] max_exec_time_array;

       // Accumulated satndard deviation
       double[] standard_deviation_array;

       // Grouped values of wallclock time together with the number of their occurences 
       long[] values_array;
       int[]  nb_values_array;
       int[]  values_indexes; // points to indexes inside values_array
   
       //  Execution time jitter
       int[] exec_time_jitter_array;


    /** Creates a new instance of Measurements */
    Measurements(int nb_loops) {

        this.nb_loops = nb_loops;

        // Instantiate empty data arrays for recording time measurements
        // of the target measured thread
        exec_time_array = new long[nb_loops];
        mean_exec_time_array = new long[nb_loops];
        delta_to_mean_exec_time_array = new int[nb_loops];
        delta_to_ref_time_array = new int[nb_loops];
        
        delta_to_previous_array = new int[nb_loops];
        mean_delta_to_previous_array = new int[nb_loops];

        min_exec_time_array = new long[nb_loops];
        max_exec_time_array = new long[nb_loops];
        
        standard_deviation_array = new double[nb_loops];

        exec_time_jitter_array = new int[nb_loops];
            
        // there will be at most nb_loops grouped values
        values_array = new long[nb_loops];
        nb_values_array = new int[nb_loops];
        values_indexes = new int[nb_loops];
    }

    // set verbosity level of measurements
    public void setVerbosity(boolean verbosity_flag) {
        verbosity = verbosity_flag;
    }

    // utility to convert a float as a two decimal string
    static private String floatAsTwoDecimalString(float aFloat) {
        long  ten_thousand_time_afloat = (long) (aFloat * 10000);
        return Long.toString(ten_thousand_time_afloat/10000)+ "," +
               Long.toString((ten_thousand_time_afloat % 10000) / 100);
    }

    // groups identical exec time value in values_array and nb_values_array
    private int groupSimilarValues() {
        int nb_values = 0;
        
        // grouping of identical values indide values_array and nb_values_array
        for (int i = 9; i < nb_loops; i++) {
            boolean found = false;
            for (int j = 0; j < nb_values; j++) {
                if (exec_time_array[i] == values_array[j]) {
                    nb_values_array[j]++;
                    found = true;
                    break;
                }
            }
            if (found == false) {
                values_array[nb_values] = exec_time_array[i];
                nb_values_array[nb_values] = 1;
                nb_values++;
            }
        }

        // indexes of groups by decreasing number of occurences
        for ( int i = 0; i < nb_values; i++) values_indexes[i]  = i;
        for (int j = 0; j < nb_values; j++) {
            for (int k = j+1; k < nb_values; k++) {
                if (nb_values_array[values_indexes[k]] >= nb_values_array[values_indexes[j]]) {
                    int tmp = values_indexes[j];
                    values_indexes[j]  = values_indexes[k];
                    values_indexes[k] = tmp;
                }
            }
        }
        
        return nb_values;
    }

    // print the deviation to previous value 
    private void computeAndPrintRawDeviationResults() {

        // deviation variables
        int nb_identical_results =0;
        int identical_results_stream_index = 0;
        int noticeable_positive_deviation = 0;   // 2 consecutive values more distant than 500 us
        int noticeable_negative_deviation = 0;   // 2 consecutive values more distant than 500 us
        int identical_results_stream_max = 0;    // max number of identical result

        // Deviations of values after 10 loops
        System.out.println("\n================================================================================");
        System.out.println("Noticeable deviations:");
        System.out.println("================================================================================");
        for (int i = 9; i < nb_loops; i++) {
            
            // find deviations streams
            if (delta_to_previous_array[i] == 0) nb_identical_results++;
            else {
                if (nb_identical_results >= identical_results_stream_max) {
                    identical_results_stream_max = nb_identical_results+ 1;
                    identical_results_stream_index = i-1;
                }
                nb_identical_results = 0;
            }
            
            if (delta_to_previous_array[i] > 500) {
                noticeable_positive_deviation++;
                if (verbosity) System.out.println("For loop " + i + ", execution time: " + exec_time_array[i] + " microseconds, deviation: " + delta_to_previous_array[i] + " microseconds");
            }
                
            if (delta_to_previous_array[i] < -500)  {
                noticeable_negative_deviation++;
                if (verbosity) System.out.println("For loop " + i + ", execution time: " + exec_time_array[i] + " microseconds, deviation: " + delta_to_previous_array[i] + " microseconds");
            }
        }

        // Summary of raw deviations
        System.out.println("\n================================================================================");
        System.out.println("Summary of deviations based on previous:");
        System.out.println("================================================================================");
        System.out.println("Nb noticeable positive deviations ( > 500 microseconds) = " + noticeable_positive_deviation);
        System.out.println("Nb noticeable negative deviations ( < -500 microseconds) = " + noticeable_negative_deviation);
        System.out.print("Nb max identical results = " + identical_results_stream_max);
        System.out.println(" for value " + exec_time_array[identical_results_stream_index]);
        System.out.println();
         
    }

    private void computeAndPrintMissedDeadlines(int[] nb_deadline_misses) {
        if (nb_deadline_misses != null) {
            boolean miss_hit = false;
            for(int i = 9; i < nb_loops; i++) {
                if (nb_deadline_misses[i] != 0) {
                    if ( !miss_hit ) {
                       miss_hit =true;
                       System.out.println("\n================================================================================");
                       System.out.println(" Deadline misses for Bench thread:" );
                       System.out.println("================================================================================");
                    }
                    
                    System.out.println("For loop " + i + ", number of deadline missed : " + nb_deadline_misses[i]);
                }
            }
        }
    }

    // computes all the exec time datas 
    private void computeRawTimeResults(long[] start_times, long[] end_times) {
        // computing exec times
        for (int i = 0; i < nb_loops; i++) 
        exec_time_array[i] = (end_times[i] - start_times[i]) / TIME_UNIT_OF_MICROSECONDS ;

        // Max, Min, and mean cpu time
        long wallclock_max_value = 0;
        long wallclock_min_value = 0;
        long wallclock_mean_value = 0;
        long total_exec_time = 0;
        long total_delta_to_previous = 0;
        long mean_delta_to_previous = 0;

        // computes data starting from loop 10 to the end
        for(int i = 9; i < nb_loops; i++) {

            // exec time for this loop
            long loop_time = exec_time_array[i];

            // Mean time
            total_exec_time += loop_time;
            wallclock_mean_value = total_exec_time / (i - 8);
            mean_exec_time_array[i] = wallclock_mean_value;

            // Deviation to mean value recording
            delta_to_mean_exec_time_array[i] = (int) (loop_time - wallclock_mean_value);

            // difference to previous loop value
            delta_to_previous_array[i] = (int) (exec_time_array[i] - exec_time_array[i - 1]);
            total_delta_to_previous += Math.abs(delta_to_previous_array[i]);
        
            // mean of difference to previous loop value
            mean_delta_to_previous = (total_delta_to_previous / (i - 8));
            mean_delta_to_previous_array[i] = (int) mean_delta_to_previous;

            // max and min values
            if (loop_time > wallclock_max_value) wallclock_max_value = loop_time;
            if ((loop_time < (wallclock_min_value - 1)) || (wallclock_min_value == 0)) {
                wallclock_min_value = loop_time;
            }
            min_exec_time_array[i] = wallclock_min_value;
            max_exec_time_array[i] = wallclock_max_value;
             
            // standard deviation until this loop
            double variance = 0;
            for (int j = 9; j < i; j++) {
                variance = variance + Math.pow((double) (exec_time_array[j]) - (double)( mean_exec_time_array[i]), 2);
            }
            variance = variance / (double)(i - 8);
            standard_deviation_array[i] = Math.sqrt(variance);

            // execution time jitter
            exec_time_jitter_array[i] = (int) (wallclock_max_value - wallclock_min_value);
            
        }
    }

    // prints in a tabed fashion the exec time raw measures
    private void printTimeMeasurementsResults(long[] starting_times, long[] ending_times) {
        System.out.println("\n================================================================================");
        System.out.println("Time measurement results:");
        System.out.println("================================================================================");
        for (int i = 0; i < nb_loops; i++) {
            System.out.print("Loop_no: " + i + " | ");
            System.out.print("exec_time: " + exec_time_array[i] + " | ");

            if (i >= 9) {
                System.out.print("start time: " + starting_times[i]  + " | ");
                System.out.print("end time: " + ending_times[i]  + " | ");
                System.out.print("mean_exec: " + mean_exec_time_array[i] + " | ");
                System.out.print("delta_previous: " + delta_to_previous_array[i] + " | ");
                System.out.print("mean_delta_previous: " + mean_delta_to_previous_array[i] + " | ");
                System.out.print("delta_mean_exec: " + delta_to_mean_exec_time_array[i]  + " | ");
                System.out.print("delta_ref: " + delta_to_ref_time_array[i]  + " | ");
                System.out.print("min_exec: " + min_exec_time_array[i] + " | ");
                System.out.print("max_exec: " + max_exec_time_array[i] + " | ");
                System.out.print("std_dev: " + floatAsTwoDecimalString((float) standard_deviation_array[i]) + " | ");
                System.out.print("exec_jitter: " + exec_time_jitter_array[i] + " | ");
            }
                
            System.out.println();
        }
    }

    // computes and prints determinism against a given ref value
    private void computeAndPrintDeterminism(int nb_values, long ref_value, String value_message) {

        // Estimation of determinism
        long total_deviation_time = 0;
        long deviation_time_less_5 = 0;
        long deviation_time_less_10 = 0;
        long deviation_time_less_20 = 0;
        long deviation_time_less_40 = 0;
        long deviation_time_less_50 = 0;
        long deviation_time_less_75 = 0;
        long deviation_time_less_100 = 0;
        
        long deviation_time_less_200 = 0;
        long deviation_time_less_300 = 0;
        long deviation_time_less_400 = 0;
        long deviation_time_less_500 = 0;
        long deviation_time_less_750 = 0;
        long deviation_time_less_1000 = 0;
        
        for (int i = 0; i < nb_values; i++) {
            long tmp_deviation = values_array[values_indexes[i]] - ref_value;
            long absolute_tmp_deviation = Math.abs(tmp_deviation);
            long tmp_time = absolute_tmp_deviation * nb_values_array[values_indexes[i]];
            
            if (i > 0) {
                if (absolute_tmp_deviation < 5) deviation_time_less_5 = deviation_time_less_5 + tmp_time;
                if (absolute_tmp_deviation < 10) deviation_time_less_10 = deviation_time_less_10 + tmp_time;
                if (absolute_tmp_deviation < 20) deviation_time_less_20 = deviation_time_less_20 + tmp_time;
                if (absolute_tmp_deviation < 40) deviation_time_less_40 = deviation_time_less_40 + tmp_time;
                if (absolute_tmp_deviation < 50) deviation_time_less_50 = deviation_time_less_50 + tmp_time;
                if (absolute_tmp_deviation < 75) deviation_time_less_75 = deviation_time_less_75 + tmp_time;
                if (absolute_tmp_deviation < 100) deviation_time_less_100 = deviation_time_less_100 + tmp_time;
                if (absolute_tmp_deviation < 200) deviation_time_less_200 = deviation_time_less_200 + tmp_time;
                if (absolute_tmp_deviation < 300) deviation_time_less_300 = deviation_time_less_300 + tmp_time;
                if (absolute_tmp_deviation < 400) deviation_time_less_400 = deviation_time_less_400 + tmp_time;
                if (absolute_tmp_deviation < 500) deviation_time_less_500 = deviation_time_less_500 + tmp_time;
                if (absolute_tmp_deviation < 750) deviation_time_less_750 = deviation_time_less_750 + tmp_time;
                if (absolute_tmp_deviation < 1000) deviation_time_less_1000 = deviation_time_less_1000 + tmp_time;
                total_deviation_time = total_deviation_time + tmp_time;
                
            }
        }
        
        // Estimation of determinism to ref_value. 
        // Decrease from 5 microseconds deviation to 1000 microseconds deviation.
        // In the ideal case, should stay 1 for all deviations slices
        float determinism;
        determinism = 1 - ((float) deviation_time_less_5 / (float) ((nb_loops- 10) * ref_value));
        System.out.println("5 microseconds determinism: " + determinism);
        determinism = 1 - ((float) deviation_time_less_10 / (float) ((nb_loops - 10) * ref_value));
        System.out.println("10 microseconds determinism: " + determinism);
        determinism = 1 - ((float) deviation_time_less_20 / (float) ((nb_loops - 10) * ref_value));
        System.out.println("20 microseconds determinism: " + determinism);
        determinism = 1 - ((float) deviation_time_less_40 / (float) ((nb_loops - 10) * ref_value));
        System.out.println("40 microseconds determinism: " + determinism);
        determinism = 1 - ((float) deviation_time_less_50 / (float) ((nb_loops - 10) * ref_value));
        System.out.println("50 microseconds determinism: " + determinism);
        determinism = 1 - ((float) deviation_time_less_75 / (float) ((nb_loops - 10) * ref_value));
        System.out.println("75 microseconds determinism: " + determinism);
        determinism = 1 - ((float) deviation_time_less_100 / (float) ((nb_loops - 10)* ref_value));
        System.out.println("100 microseconds determinism: " + determinism);
        determinism = 1 - ((float) deviation_time_less_200 / (float) ((nb_loops- 10)* ref_value));
        System.out.println("200 microseconds determinism: " + determinism);
        determinism = 1 - ((float) deviation_time_less_300 / (float) ((nb_loops- 10)* ref_value));
        System.out.println("300 microseconds determinism: " + determinism);
        determinism = 1 - ((float) deviation_time_less_400 / (float) ((nb_loops- 10)* ref_value));
        System.out.println("400 microseconds determinism: " + determinism);
        determinism = 1 - ((float) deviation_time_less_500 / (float) ((nb_loops- 10)* ref_value));
        System.out.println("500 microseconds determinism: " + determinism);
        determinism = 1 - ((float) deviation_time_less_750 / (float) ((nb_loops- 10)* ref_value));
        System.out.println("750 microseconds determinism: " + determinism);
        determinism = 1 - ((float) deviation_time_less_1000 / (float) ((nb_loops- 10)* ref_value));
        System.out.println("1000 microseconds determinism: " + determinism);
        
        determinism = 1 - ( (float) total_deviation_time / (float) ((nb_loops - 10) * ref_value));
        System.out.println("\n Overall determinism based on " + value_message + ref_value + " is: " + determinism);
        System.out.println("--------------------------------------------------------------------------------\n");
    }

    // sums up the list of deviation to ref_value, as a percentage
    private void computeAndPrintDeviations(int nb_values, long ref_value) {

        int deviation_0_to_5   = 0;
        int deviation_5_to_10  = 0;
        int deviation_10_to_20 = 0;
        int deviation_20_to_40 = 0;
        int deviation_40_to_50 = 0;
        int deviation_50_to_75 = 0;
        int deviation_75_to_100 = 0;
        int deviation_100_to_200= 0;
        int deviation_200_to_300 = 0;
        int deviation_300_to_400 = 0;
        int deviation_400_to_500 = 0;
        int deviation_500_to_750 = 0;
        int deviation_750_to_1000 = 0;
        int deviation_more_1000 = 0;

        for (int i = 0; i < nb_values; i++) {
            if (values_array[i] != ref_value) {
                if (values_array[i] >= ref_value) {
                    if (values_array[i] <=  (ref_value + 5))   deviation_0_to_5 = deviation_0_to_5 + nb_values_array[i];
                    else if (values_array[i] <=  (ref_value + 10))  deviation_5_to_10 = deviation_5_to_10 + nb_values_array[i];
                    else if (values_array[i] <=  (ref_value + 20))  deviation_10_to_20 = deviation_10_to_20 + nb_values_array[i];
                    else if (values_array[i] <=  (ref_value + 40))  deviation_20_to_40 = deviation_20_to_40 + nb_values_array[i];
                    else if (values_array[i] <=  (ref_value + 50))  deviation_40_to_50 = deviation_40_to_50 + nb_values_array[i];
                    else if (values_array[i] <=  (ref_value + 75))  deviation_50_to_75 = deviation_50_to_75 + nb_values_array[i];
                    else if (values_array[i] <=  (ref_value + 100)) deviation_75_to_100 = deviation_75_to_100 + nb_values_array[i];
                    else if (values_array[i] <=  (ref_value + 200)) deviation_100_to_200 = deviation_100_to_200 + nb_values_array[i];
                    else if (values_array[i] <=  (ref_value + 300)) deviation_200_to_300 = deviation_200_to_300 + nb_values_array[i];
                    else if (values_array[i] <=  (ref_value + 400)) deviation_300_to_400 = deviation_300_to_400 + nb_values_array[i];
                    else if (values_array[i] <=  (ref_value + 500)) deviation_400_to_500 = deviation_400_to_500 + nb_values_array[i];
                    else if (values_array[i] <=  (ref_value + 750)) deviation_500_to_750 = deviation_500_to_750 + nb_values_array[i];
                    else if (values_array[i] <=  (ref_value + 1000)) deviation_750_to_1000 = deviation_750_to_1000 + nb_values_array[i];
                    else deviation_more_1000 = deviation_more_1000 + nb_values_array[i];
                }
                else {
                    if (values_array[i] >=  (ref_value - 5))  deviation_0_to_5  = deviation_0_to_5 + nb_values_array[i];
                    else if (values_array[i] >=  (ref_value - 10))  deviation_5_to_10  = deviation_5_to_10 + nb_values_array[i];
                    else if (values_array[i] >=  (ref_value - 20))  deviation_10_to_20 = deviation_10_to_20 + nb_values_array[i];
                    else if (values_array[i] >=  (ref_value - 40))  deviation_20_to_40 = deviation_20_to_40 + nb_values_array[i];
                    else if (values_array[i] >=  (ref_value - 50))  deviation_40_to_50 = deviation_40_to_50 + nb_values_array[i];
                    else if (values_array[i] >=  (ref_value - 75))  deviation_50_to_75 = deviation_50_to_75 + nb_values_array[i];
                    else if (values_array[i] >=  (ref_value - 100))  deviation_75_to_100 = deviation_75_to_100 + nb_values_array[i];
                    else if (values_array[i] >=  (ref_value - 200)) deviation_100_to_200 = deviation_100_to_200 + nb_values_array[i];
                    else if (values_array[i] >=  (ref_value - 300)) deviation_200_to_300 = deviation_200_to_300 + nb_values_array[i];
                    else if (values_array[i] >=  (ref_value - 400)) deviation_300_to_400 = deviation_300_to_400 + nb_values_array[i];
                    else if (values_array[i] >=  (ref_value - 500)) deviation_400_to_500 = deviation_400_to_500 + nb_values_array[i];
                    else if (values_array[i] >=  (ref_value - 750)) deviation_500_to_750 = deviation_500_to_750 + nb_values_array[i];
                    else if (values_array[i] >=  (ref_value - 1000)) deviation_750_to_1000 = deviation_750_to_1000 + nb_values_array[i];
                    else deviation_more_1000 = deviation_more_1000 + nb_values_array[i];
                }
            }
        }
        
        System.out.println("Lower values below are exclusive, higher value are inclusive : ");

        System.out.println("0 to 5 microseconds deviations: " + deviation_0_to_5 + ", " + 
                           floatAsTwoDecimalString(((float) deviation_0_to_5 / (float) nb_loops ) * 100) + " %");
        System.out.println("5 to 10 microseconds deviations: " + deviation_5_to_10 + ", " +
                           floatAsTwoDecimalString(((float) deviation_5_to_10 / (float) nb_loops ) * 100) + " %");
        System.out.println("10 to 20 microseconds deviations: " + deviation_10_to_20 + ", " +
                           floatAsTwoDecimalString(((float) deviation_10_to_20 / (float) nb_loops ) * 100) + " %");
        System.out.println("20 to 40 microseconds deviations: " + deviation_20_to_40 + ", " +
                           floatAsTwoDecimalString(((float) deviation_20_to_40 / (float) nb_loops ) * 100) + " %");
        System.out.println("40 to 50 microseconds deviations: " + deviation_40_to_50 + ", " +
                           floatAsTwoDecimalString(((float) deviation_40_to_50 / (float) nb_loops ) * 100) + " %");
        System.out.println("50 to 75 microseconds deviations: " + deviation_50_to_75 + ", " +
                           floatAsTwoDecimalString(((float) deviation_50_to_75 / (float) nb_loops ) * 100) + " %");
        System.out.println("75 to 100 microseconds deviations: " + deviation_75_to_100 + ", " +
                           floatAsTwoDecimalString(((float) deviation_75_to_100 / (float) nb_loops ) * 100) + " %");
        System.out.println("100 to 200 microseconds deviations: " + deviation_100_to_200 + ", " +
                           floatAsTwoDecimalString(((float) deviation_100_to_200 / (float) nb_loops ) * 100) + " %");
        System.out.println("200 to 300 microseconds deviations: " + deviation_200_to_300 + ", " +
                           floatAsTwoDecimalString(((float) deviation_200_to_300 / (float) nb_loops ) * 100) + " %");
        System.out.println("300 to 400 microseconds deviations: " + deviation_300_to_400 + ", " +
                           floatAsTwoDecimalString(((float) deviation_300_to_400 / (float) nb_loops ) * 100) + " %");
        System.out.println("400 to 500 microseconds deviations: " + deviation_400_to_500 + ", " +
                           floatAsTwoDecimalString(((float) deviation_400_to_500 / (float) nb_loops ) * 100) + " %");
        System.out.println("500 to 750 microseconds deviations: " + deviation_500_to_750 + ", " +
                           floatAsTwoDecimalString(((float) deviation_500_to_750 / (float) nb_loops ) * 100) + " %");
        System.out.println("750 to 1000 microseconds deviations: " + deviation_750_to_1000 + ", " +
                           floatAsTwoDecimalString(((float) deviation_750_to_1000 / (float) nb_loops ) * 100) + " %");
        System.out.println("More than 1000 microseconds deviations: " + deviation_more_1000 + ", " +
                           floatAsTwoDecimalString(((float) deviation_more_1000 / (float) nb_loops ) * 100) + " %");
    }

    // summarize the recorded values, grouped by
    //  - decreasing order of occurence
    //  - and decreasing deviation
    private void computeAndPrintOrderedValues(int nb_values, long ref_value, long mean_value) {

        System.out.println("\n================================================================================");
        System.out.println("Ordered list of values:" );
        System.out.println("================================================================================");
        int index = 0;
        while (index < nb_values) {
            int tmp_nb_values = nb_values_array[values_indexes[index]];
                
            // find a group of occurences
            int tmp_index = index;
            long subgroup_max_value = values_array[values_indexes[tmp_index]];
            while ((tmp_index < nb_values) &&
                   (nb_values_array[values_indexes[tmp_index]] == tmp_nb_values)) {
                tmp_index++;
            }
                
            // iterates on this group to print results ordered by decreasing order of deviation
            long last_max_value = 0;
            for (int i = index; i < tmp_index; i++) {
                long current_max_value = 0;
                int current_index = 0;
                int current_dev = 0;
                    
                for (int j = index; j < tmp_index; j++) {
                    long value = values_array[values_indexes[j]];
                    if (i == index) {
                        // find greatest value of this subgroup
                        if (value > current_max_value) {
                            current_index = j;
                            current_max_value = value;
                        }
                    }
                    else if ((value > current_max_value) &&
                             (value < last_max_value)) {
                        // this one may be the greatest still smaller than the last greatest
                        current_index = j;
                        current_max_value = value;
                    }
                }
                System.out.print("   Execution time " + current_max_value);
                System.out.print(" occurred " + nb_values_array[values_indexes[current_index]] + " times, ");
                System.out.print(" ref deviation is " + (int) (current_max_value - ref_value) + " microseconds, ");
                System.out.print(" mean deviation is " + (int) (current_max_value - mean_value) + " microseconds");
                System.out.println();

                last_max_value = current_max_value;
            }
                
            System.out.println();
            index = tmp_index;
        }
    }

    // big method to compute and print results based on recorded measurements
    void computeAndPrintStatResults(long[] starting_times, long[] ending_times, int[] nb_deadline_misses) {

        computeRawTimeResults(starting_times, ending_times);

        // overall values
        long overall_mean_exec_time = mean_exec_time_array[nb_loops - 1];
        long overall_mean_delta_to_previous = mean_delta_to_previous_array[nb_loops - 1];
        long overall_best_execution_time = min_exec_time_array[nb_loops - 1];
        long overall_worst_execution_time = max_exec_time_array[nb_loops - 1];
        int  overall_exec_time_jitter = exec_time_jitter_array[nb_loops - 1];
        double overall_standard_deviation = standard_deviation_array[nb_loops - 1];

        // grouping of exec time values 
        int nb_values = groupSimilarValues();

        // most frequent value is the first referenced from values_indexes
        long most_frequent_value = values_array[values_indexes[0]];
         
        // computing delta to ref value
        for (int i = 9; i < nb_loops; i++)
            delta_to_ref_time_array[i] = (int) (exec_time_array[i] - most_frequent_value);

        // printing of raw temporal measurements and ordered list of values
        if (verbosity == true)
            printTimeMeasurementsResults(starting_times, ending_times);
        
        // raw deviation results
        computeAndPrintRawDeviationResults();

        // deviations to ref and min values 
        System.out.println("\n======================================================================");
        System.out.println("Percentage report of deviations:");
        System.out.println("======================================================================\n");
        System.out.print("--------------- deviations based on most frequent value ");
        System.out.println(most_frequent_value + " ---------------");
        computeAndPrintDeviations(nb_values, most_frequent_value);
        System.out.print("\n--------------- deviations based on mean execution time ");
        System.out.println(overall_mean_exec_time + " ---------------");
        computeAndPrintDeviations(nb_values, overall_mean_exec_time);

        // sums up the list of exec times in an ordered fashion
        if (verbosity == true) {
            // Dont show deadline misses by default (too verbose)
            // computeAndPrintMissedDeadlines(nb_deadline_misses);
           computeAndPrintOrderedValues(nb_values, most_frequent_value, overall_mean_exec_time);
        }
        // Summary of results
        System.out.println("\n======================================================================");
        System.out.println("Summary of results:");
        System.out.println("========================================================================");
        
        System.out.println("Mean execution time: " + overall_mean_exec_time + " microseconds");
        System.out.println("Best execution time: " + overall_best_execution_time + " microseconds");
        System.out.println("Worst execution time: " + overall_worst_execution_time + " microseconds");
        System.out.print("Most frequent execution time: " + values_array[values_indexes[0]] + " microseconds, " );
        System.out.println(nb_values_array[values_indexes[0]] + " occurences");
        System.out.println("Execution time jitter: " + overall_exec_time_jitter);
        System.out.println("Standard deviation: " + floatAsTwoDecimalString((float) overall_standard_deviation));
        System.out.println("\n");
        
        }
}


